import sys

import datetime

from mlpug.trainers.callbacks.callback import Callback
from mlpug.utils import get_value_at, is_chunkable

import basics.base_utils as _


class LogProgress(Callback):

    def __init__(self, log_period=200, set_names=None, batch_level=True, name="LogProgress"):
        super(LogProgress, self).__init__(name=name)

        self.log_period = log_period
        self.set_names = set_names or ["training"]
        self.batch_level = batch_level

    def on_batch_training_completed(self, training_batch, logs):
        if not self.batch_level:
            return True

        success = True
        batch_step = logs["batch_step"]
        if batch_step == 0 or batch_step % self.log_period == 0:
            eta = self._calc_eta(logs)
            average_duration = self._get_average_batch_duration(logs)
            sys.stdout.write('Epoch {:d}/{:d} - ETA: {:s}\tBatch {:d}/{:d} '
                             'Average batch training time {:s}\n'.format(logs["epoch"],
                                                                           logs["final_epoch"],
                                                                           eta,
                                                                           logs["batch_step"],
                                                                           logs["final_batch_step"],
                                                                           average_duration))
            sys.stdout.write('Batch:\n')
            for set_name in self.set_names:
                metrics_log = self._create_set_metrics_log_for(set_name, logs, mean_metrics=False)
                if metrics_log is None:
                    success = False
                    continue

                sys.stdout.write(f'{set_name}:\t{metrics_log}.\n')

            sys.stdout.write('\nMoving average:\n')
            for set_name in self.set_names:
                metrics_log = self._create_set_metrics_log_for(set_name, logs, mean_metrics=True)
                if metrics_log is None:
                    success = False
                    continue

                sys.stdout.write(f'{set_name}:\t{metrics_log}.\n')

            sys.stdout.write(f'\n\n')

        return success

    def on_epoch_completed(self, logs):
        duration = self._get_epoch_duration(logs)
        sys.stdout.write('Epoch {:d}/{:d}\tREADY - Duration {:s}\n'.format(logs["epoch"],
                                                                           logs["final_epoch"],
                                                                           duration))
        success = True
        sys.stdout.write('Average:\n')
        for set_name in self.set_names:
            metrics_log = self._create_set_metrics_log_for(set_name, logs, mean_metrics=True)
            if metrics_log is None:
                success = False
                continue

            sys.stdout.write(f'{set_name}:\t{metrics_log}.\n')

        sys.stdout.write(f'\n')

        return success

    def _calc_eta(self, logs):

        eta_str = None
        try:
            average_batch_duration = logs["duration"]["mean"]["batch"]
            if average_batch_duration > 0:
                batch_step = logs["batch_step"]
                final_batch_step = logs["final_batch_step"]
                num_batches_to_go = final_batch_step - batch_step + 1

                eta_seconds = int(round(average_batch_duration * num_batches_to_go))

                eta_str = str(datetime.timedelta(seconds=eta_seconds))
            else:
                eta_str = "[UNKNOWN]"
        except Exception as e:
            _.log_exception(self._log, "Unable to calculate epoch ETA", e)

        return eta_str

    def _get_average_batch_duration(self, logs):
        duration_str = None
        try:
            duration = logs["duration"]["mean"]["batch"]
            if duration > 0.0:
                duration = int(duration*1000)
                duration_str = f"{duration}ms"
            else:
                duration_str = "[UNKNOWN]"
        except Exception as e:
            _.log_exception(self._log, "Unable to get average batch duration", e)

        return duration_str

    def _get_epoch_duration(self, logs):
        duration_str = None
        try:
            epoch_duration = int(round(logs["duration"]["epoch"]))
            duration_str = str(datetime.timedelta(seconds=epoch_duration))
        except Exception as e:
            _.log_exception(self._log, "Unable to get epoch duration", e)

        return duration_str

    def _create_set_metrics_log_for(self, set_name, logs, mean_metrics):
        key_path = set_name
        if mean_metrics:
            key_path += '.mean'
        metrics = get_value_at(key_path, logs)

        metrics_log = self._create_log_for(metrics)
        if metrics_log is None:
            self._log.error(f'No {"mean" if mean_metrics else ""} metrics data available for {key_path}, '
                            f'unable to create log for these set metrics')
            return None

        return metrics_log

    def _create_log_for(self, metrics):
        if not _.is_dict(metrics):
            return None

        metrics = metrics.copy()
        if 'mean' in metrics:
            del metrics['mean']

        num_metrics = len(metrics)

        log_format = "{:s} {:.3f}"
        log = ""
        for c, (metric, value) in enumerate(metrics.items()):
            if metric == "mean":
                continue

            log += log_format.format(metric, value)
            if c < num_metrics - 1:
                log += ', '

        return log


class BatchSizeLogger(Callback):

    def __init__(self, batch_dimension=1, name="BatchSizeLogger"):
        super().__init__(name=name)

        self._batch_dimension = batch_dimension

    def on_batch_training_start(self, training_batch, logs):
        """

        :param training_batch:
        :param logs:

        :return: success (True or False)
        """

        logs['batch_size'] = len(training_batch) if is_chunkable(training_batch) else \
            training_batch[0].size(self._batch_dimension)

        return True




import math

import numpy as np

from mlpug.trainers.callbacks.callback import Callback
from mlpug.utils import has_key, SlidingWindow

import basics.base_utils as _


class MetricsLoggerBase(Callback):

    def __init__(self,
                 dataset_name,
                 metrics,
                 batch_averaging_window=None,
                 name="MetricsLoggerBase"):
        super().__init__(name=f"{dataset_name} dataset {name}")

        self._metrics = metrics

        self._dataset_name = dataset_name

        self._batch_averaging_window = batch_averaging_window

        self._metric_averages = None
        self._metric_windows = None

    def get_state(self):
        """

        :return: state, success (True or False)
        """
        return {
                    "batch_averaging_window": self._batch_averaging_window,
                    "metric_averages": self._metric_averages,
                    "metric_windows": {metric_name: window.get_state()
                                       for metric_name, window in self._metric_windows.items()}
               }, True

    def set_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """

        if not self._check_state(state):
            self._log.error('Given state is invalid, unable to set state')
            return False

        def create_sliding_window(ws):
            return SlidingWindow(ws["length"], init_window_values=ws["window"])

        self._batch_averaging_window = state["batch_averaging_window"]
        self._metric_averages = state["metric_averages"]
        self._metric_windows = {metric_name: create_sliding_window(window_state)
                                for metric_name, window_state in state["metric_windows"].items()}

        return True

    def on_epoch_start(self, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if self._dataset_name not in logs:
            logs[self._dataset_name] = {}

        if 'mean' not in logs[self._dataset_name]:
            logs[self._dataset_name]['mean'] = self._metric_averages or {}

        return True

    def on_batch_training_completed(self, dataset_batch, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if not self._calc_metrics_on(dataset_batch, logs[self._dataset_name]):
            return False

        metric_names = self._metrics.keys()
        if not self._update_metrics_windows_for(metric_names, logs[self._dataset_name]):
            return False

        if not self._calc_metric_window_averages(metric_names, logs[self._dataset_name]["mean"]):
            return False

        return True

    def _setup_metrics_averaging(self, window_length, reset=False):
        if reset:
            self._metric_averages = {}
            self._metric_windows = {}

        self._metric_averages = self._metric_averages or {}
        self._metric_windows = self._metric_windows or {}

        for metric_name in self._metrics.keys():
            if metric_name not in self._metric_windows:
                self._metric_windows[metric_name] = SlidingWindow(window_length)
                self._metric_averages[metric_name] = None

        return True

    def _calc_metrics_on(self, batch, metrics):
        success = True

        for metric_name, metric_func in self._metrics.items():
            try:
                metrics[metric_name] = metric_func(batch)
            except Exception as e:
                _.log_exception(self._log, f"Exception occurred calculating {metric_name} on "
                                           f"{self._dataset_name} batch", e)
                success = False

        return success

    def _update_metrics_windows_for(self, metric_names, batch_metric_values):
        success = True

        for metric_name in metric_names:
            try:
                metric_value = batch_metric_values[metric_name]
                self._metric_windows[metric_name].slide(metric_value)
            except Exception as e:
                _.log_exception(self._log, f"Exception occurred updating sliding averaging window for "
                                           f"{self._dataset_name} batch {metric_name}", e)
                success = False

        return success

    def _calc_metric_window_averages(self, metric_names, metrics_averages):
        success = True

        for metric_name in metric_names:
            try:
                window_data = self._metric_windows[metric_name].window
                metrics_averages[metric_name] = np.mean(np.array(window_data))
            except Exception as e:
                _.log_exception(self._log, f"Exception occurred calculating sliding window average for "
                                           f"{self._dataset_name} batch {metric_name}", e)
                success = False

        return success

    def _check_state(self, state):
        state_attributes = ['batch_averaging_window', 'metric_averages', 'metric_windows']

        for attr in state_attributes:
            if not has_key(state, attr):
                self._log.error(f"Given state does not have a value for {attr}, state is invalid")
                return False

        return True


class TrainingMetricsLogger(MetricsLoggerBase):

    def __init__(self, metrics, batch_averaging_window, name="MetricsLogger"):
        super().__init__(dataset_name='training',
                         metrics=metrics,
                         batch_averaging_window=batch_averaging_window,
                         name=name)

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if self._batch_averaging_window is None:
            self._log.debug(f"Set batch_averaging_window to number of batches per epoch : {num_batches_per_epoch}")
            self._batch_averaging_window = num_batches_per_epoch

        if self._batch_averaging_window == math.inf:
            self._log.error("The batch average window is infinite, unable to calculate window average. "
                            "Please set a finite batch_averaging_window during construction of this "
                            "TrainingMetricsLogger.")
            self._valid = False
            return False

        return self._setup_metrics_averaging(self._batch_averaging_window)


class TestMetricsLogger(MetricsLoggerBase):

    def __init__(self,
                 dataset,
                 dataset_name,
                 metrics,
                 batch_level=True,
                 batch_assessment_period=1,
                 batch_averaging_window=None,
                 name="MetricsLogger"):
        super().__init__(dataset_name=dataset_name,
                         metrics=metrics,
                         batch_averaging_window=batch_averaging_window,
                         name=name)

        self._dataset = dataset

        self._batch_level = batch_level
        self._batch_assessment_period = batch_assessment_period
        self._batch_averaging_window = batch_averaging_window

        self._dataset_iterator = None

        self._validate()

    def get_state(self):
        if self._batch_level:
            return super().get_state()
        else:
            return None, True

    def set_state(self, state):
        if self._batch_level:
            return super().set_state(state)
        else:
            return True

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        return self._setup_metrics_averaging(self._batch_averaging_window)

    def on_epoch_start(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """
        success = super().on_epoch_start(logs)

        if not success:
            self._log.error('A problem occurred, will not continue executing this hook')
            return success

        if self._batch_level and self._dataset_iterator is None:
            self._dataset_iterator = iter(self._dataset)

        return True

    def on_batch_training_completed(self, training_batch, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if not self._batch_level:
            return True

        training_iter = logs['global_iter']

        if training_iter % self._batch_assessment_period != 0:
            return True

        try:
            dataset_batch = next(self._dataset_iterator)
        except StopIteration:
            self._log.debug(f"[Iteration {training_iter}] Rewinding the {self._dataset_name} data set")
            self._dataset_iterator = iter(self._dataset)
            dataset_batch = next(self._dataset_iterator)

        return super().on_batch_training_completed(dataset_batch, logs)

    def on_epoch_completed(self, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if self._batch_level:
            return True

        self._dataset_iterator = iter(self._dataset)

        self._setup_metrics_averaging(self._batch_averaging_window, reset=True)

        metric_names = self._metrics.keys()
        # Loop over all dataset batches, this will fill the sliding windows, finally calculate the averages
        for dataset_batch in self._dataset_iterator:
            batch_metrics = {}
            if not self._calc_metrics_on(dataset_batch, batch_metrics):
                return False

            if not self._update_metrics_windows_for(metric_names, batch_metrics):
                return False

        return self._calc_metric_window_averages(metric_names, logs[self._dataset_name]["mean"])

    def _validate(self):
        self._valid = True

        if not hasattr(self._dataset, '__iter__'):
            self._log.error(f"The given dataset {str(self._dataset)} is not iterable, the {self} will not function")
            self._valid = False

        if self._batch_averaging_window is None:
            try:
                self._batch_averaging_window = len(self._dataset)
                if self._batch_averaging_window == math.inf:
                    self._log.error(f"The batch average window of the {self._dataset_name} data set is infinite, "
                                    f"unable to calculate window average. Please set a finite batch_averaging_window "
                                    f"during construction of this TestMetricsLogger.")
                    self._valid = False
                else:
                    self._log.debug(f"Set batch averaging window to number of batches in "
                                    f"{self._dataset_name} data set : {self._batch_averaging_window}")
            except Exception as e:
                _.log_exception(self._log, f"Unable to assess data set length to set the batch averaging window, "
                                           f"{self} will not function", e)
                self._valid = False








from .basic import *
from .callback import *
from .checkpoint_manager import CheckpointManager
from .lr_scheduler_wrapper import LRSchedulerWrapper
from .metrics_logger import TrainingMetricsLogger, TestMetricsLogger
from .tensorboard import Tensorboard, AutoTensorboard






import os
import basics.base_utils as _
from basics.logging import get_logger

logger = get_logger(os.path.basename(__file__))


def convert_to_dict(type, components):
    if _.is_sequence(components):
        if len(components) == 1:
            components = {
                type: components[0]
            }
        elif len(components) > 1:
            components = {f"{type}_{i}": component for i, component in enumerate(components)}
    elif not _.is_dict(components):
        # Assuming single optimizer
        components = {
            type: components
        }

    return components


def get_value_at(key_path, nested_data, default=None, warn_on_failure=True):
    """
    Safe way to get value from nested data structure (e.g. nested dict) based on a key path

    TODO migrate to PyBase

    :param key_path:
    :param nested_data:
    :param default:
    :param warn_on_failure:

    :return:
    """
    keys = key_path.split(".")
    value = nested_data
    for key in keys:
        if has_key(value, key):
            value = value[key]
        else:
            if warn_on_failure:
                logger.warn(f"Key path {key_path} not found in given data")
            value = None
            break

    if value is None:
        value = default

    return value


def has_key(o, key):
    """
    TODO migrate to PyBase
    :param o:
    :param key:
    :return:
    """
    return hasattr(o, '__iter__') and (key in o)


def is_chunkable(batch):
    return (batch is not None) and \
           hasattr(batch, "__len__") and callable(batch.__len__) and \
           hasattr(batch, "__getitem__") and callable(batch.__getitem__)



import sys
import abc
import math
import time

from functools import reduce

from basics.base import Base
import basics.base_utils as _

from mlpug.utils import convert_to_dict, get_value_at, has_key, is_chunkable

from mlpug.mlpug_exceptions import TrainerInvalidException, BatchNotChunkableException


class TrainingManager(Base, metaclass=abc.ABCMeta):

    def __init__(self,
                 trainer,
                 training_dataset,
                 num_epochs=50,
                 num_batches_per_epoch=None,
                 callbacks=None,
                 experiment_data=None):
        """
        Implements a simple training loop in the `_train` method. Although simple in nature, the use of callbacks
        (see the `Callback` class), provide the means to monitor, control and customize the training.

        During each epoch a new `logs` dict is created that is passed from callback event to callback event during
        batch-wise training. Before calling the given `Trainer` instance to train on a batch, `training_settings` are
        retrieved from the `logs` dict, if available. These `training_settings` are then passed to the trainer `train_on`
        method. Also see the `Trainer` class.

        The provided `training_dataset` should be iterable, and provide a batch per iteration.

        :param trainer:
        :param training_dataset:
        :param num_epochs:
        :param num_batches_per_epoch:
        :param callbacks:
        :param experiment_data: Data object that will be added to the training manager state.
                                This could contain additional training/model experiment data

        """
        super(TrainingManager, self).__init__()

        self.trainer = trainer

        # Must implement __len__, __iter__
        self.training_dataset = training_dataset

        self.num_epochs = num_epochs
        self.num_batches_per_epoch = num_batches_per_epoch

        self.global_iter = 0
        self.batch_step = 0
        self.epoch = 0

        self.callbacks = callbacks or []
        self.callbacks_map = None

        self.experiment_data = experiment_data

        self.init_logs = None
        self.logs = {}

        self._stop_training = False

        self._callback_calls_successful = True

        if not self._validate():
            return

        self._assess_num_batches_per_epoch()
        if not self._call_callbacks("set_training_manager", self):
            self._callback_calls_successful = False
            self._log.error("One or more issues occurred while providing this training manager instance "
                            "to the provided callbacks")

    def get_trainer(self):
        return self.trainer

    def start_training(self):
        if not self.instance_valid():
            self._log.error('TrainingManager is not valid, unable to start training')
            return

        try:
            self._train()
        except KeyboardInterrupt:
            sys.stdout.write("\n\n")
            self._log.info('Training process interrupted by you ... 🤷🏻‍♂️\n')
            # TODO : deal with the distributed training case, wait to return until
            #  all workers have stopped training
            self.stop_training()

    def stop_training(self):
        self._stop_training = True

    def get_state(self):
        """

        :return: state, success
        """

        if not self.instance_valid():
            self._log.error('TrainingManager is not valid, unable to get training manager state')
            return None, False

        state = {
            "manager": {
                "epoch": self.epoch,
                "batch_step": self.batch_step,
                "global_iter": self.global_iter,
                "logs": self.logs
            },

            "callbacks": {},

            "experiment_data": self.experiment_data
        }

        success = True
        for callback in self.callbacks:
            cb_name = None
            try:
                cb_name = str(callback)
                cb_state, cb_success = callback.get_state()

                success &= cb_success

                state["callbacks"][cb_name] = cb_state
            except Exception as e:
                _.log_exception(self._log, f"Failed to get state of callback {cb_name}, "
                                           f"unable to add callback state to Training Manager state", e)
                success = False

        state["trainer"], get_trainer_state_success = self.trainer.get_state()
        success &= get_trainer_state_success

        return state, success

    def set_state(self, state):
        """
        WARNING: the start batch_step in the state only controls how many batches will be trained on in the
                 the current epoch, it does not control the exact batch data that will be used for training.

        :param state:
        :return: success (True or False)
        """
        if not self.instance_valid():
            self._log.error('TrainingManager is not valid, unable to set training manager state')
            return False

        if not self._check_state(state):
            self._log.error("Invalid state object, unable to set state")
            return False

        self.epoch = state["manager"]["epoch"]
        self.batch_step = state["manager"]["batch_step"]
        self.global_iter = state["manager"]["global_iter"]
        self.init_logs = state["manager"]["logs"]

        success = True
        try:
            callbacks_map = {str(callback): callback for callback in self.callbacks}

            callbacks = state["callbacks"] or {}
            for cb_name, cb_state in callbacks.items():
                if cb_name not in callbacks_map:
                    self._log.warn(f"Callback {cb_name} not given, unable to set callback state, skipping ...")
                    continue

                callback = callbacks_map[cb_name]
                try:
                    success &= callback.set_state(cb_state)
                except Exception as e:
                    _.log_exception(self._log, f"Unable to set state for callback {cb_name}", e)
                    success = False

        except Exception as e:
            _.log_exception(self._log, f"Unable to create callbacks map, unable to set callback state", e)
            success = False

        try:
            success &= self.trainer.set_state(state["trainer"])
        except Exception as e:
            _.log_exception(self._log, f"Unable to set state for trainer", e)
            success = False

        return success

    def _check_state(self, state):
        state_attributes = ['manager',
                            'manager.epoch',
                            'manager.batch_step',
                            'manager.global_iter',
                            'manager.logs',
                            'callbacks',
                            'trainer']

        for attr in state_attributes:
            v = get_value_at(attr, state, warn_on_failure=False)
            if v is None:
                self._log.error(f"Given state does not have a value for {attr}, state is invalid")
                return False

        return True

    def _train(self):
        training_stopped_early = False
        epoch_stopped_early = False
        training_stopped_on_error = False

        cb_calls_success = self._callback_calls_successful

        cb_calls_success &= self._call_callbacks('on_training_start',
                                                 self.num_epochs,
                                                 self.num_batches_per_epoch,
                                                 self.epoch,
                                                 self.batch_step,
                                                 self.global_iter)

        for epoch in range(self.epoch, self.num_epochs):
            self.epoch = epoch

            if self._stop_training:
                training_stopped_early = True
                self._log.warn("Training stopped early")
                break

            if self.init_logs is not None:
                self._log.debug("Setting provided initial logs object ...")
                self.logs = self.init_logs

                # override log state for certain parameters given at construction
                self.logs["final_epoch"] = self.num_epochs-1
                self.logs["final_batch_step"] = self.num_batches_per_epoch - 1
                self.init_logs = None
            else:
                previous_mean_batch_training_duration = get_value_at('duration.mean.batch', self.logs, default=0)
                self.logs = {
                    "epoch": epoch,
                    "final_epoch": self.num_epochs-1,
                    "batch_step": self.batch_step,
                    "final_batch_step": self.num_batches_per_epoch-1,
                    "global_iter": self.global_iter,
                    "training": {
                        "mean": {
                            "loss": 0
                        }
                    },
                    "duration": {
                        "batch": None,
                        "epoch": 0,
                        "mean": {
                            "batch": previous_mean_batch_training_duration
                        }
                    },
                    "cb_calls_success": cb_calls_success
                }

            # It's a bit shorter
            logs = self.logs

            logs["cb_calls_success"] &= self._call_callbacks('on_epoch_start', logs)

            epoch_stopped_early = False

            summed_loss = 0
            num_batches_processed = 0
            for training_batch in iter(self.training_dataset):
                batch_training_start_time = time.time()

                if self._stop_training:
                    # Epoch was not finished yet
                    epoch_stopped_early = True
                    self._log.warn("Training epoch stopped early")
                    break

                logs["batch_step"] = self.batch_step
                logs["global_iter"] = self.global_iter
                logs["cb_calls_success"] &= self._call_callbacks('on_batch_training_start', training_batch, logs)

                try:
                    training_settings = logs["training_settings"] if has_key(logs, "training_settings") else None
                    logs["training"]["loss"], logs["auxiliary_results"] = self.trainer.train_on(training_batch,
                                                                                                training_settings)

                    num_batches_processed += 1
                    summed_loss += logs["training"]["loss"]
                    logs["training"]["mean"]["loss"] = summed_loss/num_batches_processed
                except Exception as e:
                    if isinstance(e, TrainerInvalidException):
                        err_msg = f"Trainer {self.trainer} is misconfigured, unable to train on batch, " \
                                  f"will stop training ..."
                    else:
                        err_msg = f"Exception occurred while calling {self.trainer} to train on batch, " \
                                  f"will stop training ..."

                    _.log_exception(self._log, err_msg, e)

                    logs["cb_calls_success"] &= self._call_callbacks('on_batch_training_failed', e, logs)
                    training_stopped_on_error = True

                if training_stopped_on_error:
                    break

                logs["cb_calls_success"] &= self._call_callbacks('on_batch_training_completed', training_batch, logs)

                self.batch_step += 1
                self.global_iter += 1

                batch_training_end_time = time.time()

                self.logs["duration"]["batch"] = batch_training_end_time - batch_training_start_time
                self.logs["duration"]["epoch"] += self.logs["duration"]["batch"]
                self.logs["duration"]["mean"]["batch"] = self.logs["duration"]["epoch"]/self.batch_step

                # Situation 1: When start_batch given was > 0, go to next epoch when epoch finished
                # Situation 2: When num_batches_per_epoch was given to simulates epochs in an
                #              infinite batch streaming situation
                if self.batch_step >= self.num_batches_per_epoch:
                    break

            if not (epoch_stopped_early or training_stopped_on_error):
                logs["cb_calls_success"] &= self._call_callbacks('on_epoch_completed', logs)

            cb_calls_success &= logs["cb_calls_success"]

            if training_stopped_on_error:
                break

            self.batch_step = 0

        cb_calls_success &= self._call_callbacks('on_training_ended',
                                                 stopped_early=epoch_stopped_early or training_stopped_early,
                                                 stopped_on_error=training_stopped_on_error,
                                                 callback_calls_success=cb_calls_success)

        if not training_stopped_on_error:
            if cb_calls_success:
                self._log.info("Training completed. All good! ❤️")
            else:
                self._log.warn("Training completed, but one or more callback errors occurred. "
                               "What's up with that? 🧐")
        else:
            self._log.warn("Training stopped on error. 🥺️")

    def _call_callbacks(self, event, *args, **kwargs):
        success = True
        for callback in self.callbacks:
            func_name = event
            try:
                func = getattr(callback, func_name)
                func_success = func(*args, **kwargs)
                if not _.is_bool(func_success):
                    self._log.warn(f"Success value returned by hook {func_name} by callback {callback} "
                                   f"is not boolean ({func_success}), please correct this."
                                   f"Call to hook interpreted a successful, continuing ...")
                    func_success = True

                success &= func_success
            except Exception as e:
                _.log_exception(self._log, f"Call {func_name} failed for callback {callback}", e)
                success = False

        return success

    def _assess_num_batches_per_epoch(self):
        if self.num_batches_per_epoch:
            # num_batches_per_epoch given, nothing to do
            return

        self.num_batches_per_epoch = math.inf
        err_msg = "Evaluation of training data set length failed, number of batches per epoch is " \
                  "set to Infinite. If this is not what you want, either provide the num_batches_per_epoch " \
                  "argument during construction of the TrainingManager, or ensure that the training_dataset " \
                  "given can provide the number of batches by providing an implementation of __len__ that " \
                  "does not fail."
        try:
            if hasattr(self.training_dataset, '__len__'):
                self.num_batches_per_epoch = len(self.training_dataset)
                self._log.debug(f"Number of batches per epoch : {self.num_batches_per_epoch}")
            else:
                self._log.warn(err_msg)
        except Exception as e:
            _.log_exception(self._log, err_msg, e)

    def _validate(self):
        # TODO
        self._valid = True

        return self._valid


class TrainerBase(Base, metaclass=abc.ABCMeta):

    def __init__(self, model_components, optimizers):
        """

        :param model_components: dict with model components
        :param optimizers: dict with optimizers

        """
        super(TrainerBase, self).__init__()

        self.model_components = model_components
        self.optimizers = optimizers

        self._validate()

    def get_state(self):
        """

        :return: state, success
        """

        model_components_state, mcs_success = self.get_model_components_state()
        optimizers_state, os_success = self.get_optimizers_state()

        return {
            "model_components": model_components_state,
            "optimizers": optimizers_state,
        }, mcs_success & os_success

    def set_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """
        if not self.instance_valid():
            self._log.error('Trainer is not valid, unable to set trainer state')
            return False

        if not self._check_state(state):
            self._log.error("Invalid state object, unable to set state")
            return False

        success = self.set_model_components_state(state["model_components"])
        success = self.set_optimizers_state(state["optimizers"]) & success

        return success

    def get_optimizers(self):
        return self.optimizers

    def get_optimizer(self, name):
        return get_value_at(name, self.get_optimizers())

    def get_model_components(self):
        return self.model_components

    def get_model_component(self, name):
        return get_value_at(name, self.get_model_components())

    def get_model_components_state(self):
        """

        :return: state, success (True or False)
        """
        state = {}

        model_components = self.get_model_components()

        success = True
        for name, model in model_components.items():
            try:
                state[name] = self._get_model_state(model, name)
            except Exception as e:
                _.log_exception(self._log, f"Unable to get state for model {name}", e)
                success = False

        return state, success

    def set_model_components_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """
        if not _.is_callable(getattr(state, 'items', None)):
            self._log.error("State is invalid, unable to set model components state")
            return False

        success = True
        for name, model_state in state.items():
            model = self.get_model_component(name)
            if model is None:
                self._log.error(f"No {name} model not found, unable to set state")
                success = False
                continue

            try:
                self._set_model_state(model, model_state, name)
            except Exception as e:
                _.log_exception(self._log, f"Unable to set state for model {name}", e)
                success = False

        return success

    def get_optimizers_state(self):
        """

        :return: state, success (True of False)
        """
        state = {}

        optimizers = self.get_optimizers()

        success = True
        for name, optimizer in optimizers.items():
            try:
                state[name] = self._get_optimizer_state(optimizer, name)
            except Exception as e:
                _.log_exception(self._log, f"Unable to get state for optimizer {name}", e)
                success = False

        return state, success

    def set_optimizers_state(self, state):
        """

        :param state:
        :return: success (True, False)
        """
        if not _.is_callable(getattr(state, 'items', None)):
            self._log.error("State is invalid, unable to set optimizers state")
            return False

        success = True
        for name, optimizer_state in state.items():
            optimizer = self.get_optimizer(name)
            if optimizer is None:
                self._log.error(f"No {name} optmizer not found, unable to set state")
                success = False
                continue

            try:
                self._set_optimizer_state(optimizer, optimizer_state, name)
            except Exception as e:
                _.log_exception(self._log, f"Unable to set state for optimizer {name}", e)
                success = False

        return success

    @abc.abstractmethod
    def train_on(self, batch_data, training_settings=None):
        """
        Use batch_data to perform a training iteration

        :param batch_data: batch_data object (e.g. dict, list, tuple)
        :param training_settings: optional training_settings object (usually dict)

        :return: loss, auxiliary_results

        loss : number (e.g. float)
        auxiliary_results : can be anything, e.g dict or list with values or data items
        """
        pass

    def evaluate_loss(self, batch_data, inference_mode, evaluate_settings=None):
        """

        :param batch_data: batch_data object to evaluate loss on (e.g. dict, list, tuple)
        :param inference_mode: If True the loss will be evaluated in inference mode (e.g. no Dropout).
                               If False the loss will be evaluated in training mode
        :param evaluate_settings: optional evaluate_settings object (usually dict)

        :return: loss, auxiliary_results

        loss : Tensor
        auxiliary_results : can be anything, e.g dict or list with values or data items
        """

        self._activate_inference_mode(inference_mode)

        return self._evaluate_loss(batch_data, evaluate_settings)

    @abc.abstractmethod
    def _evaluate_loss(self, batch_data, evaluate_settings=None):
        """

        :param batch_data: batch_data object to evaluate loss on (e.g. dict, list, tuple)
        :param evaluate_settings: optional evaluate_settings object (usually dict)

        :return: loss, auxiliary_results

        loss : Tensor
        auxiliary_results : can be anything, e.g dict or list with values or data items
        """

        pass

    @abc.abstractmethod
    def _activate_inference_mode(self, inference_mode):
        pass

    @abc.abstractmethod
    def _get_model_state(self, model, model_name=None):
        pass

    @abc.abstractmethod
    def _get_optimizer_state(self, optimizer, optimizer_name=None):
        pass

    @abc.abstractmethod
    def _set_model_state(self, model, state, model_name=None):
        pass

    @abc.abstractmethod
    def _set_optimizer_state(self, optimizer, state, optimizer_name):
        pass

    def _check_state(self, state):
        state_attributes = ['model_components', 'optimizers']

        for attr in state_attributes:
            v = get_value_at(attr, state, warn_on_failure=False)
            if v is None:
                self._log.error(f"Given state does not have a value for {attr}, state is invalid")
                return False

        return True

    def _validate(self):
        # TODO
        self._valid = True

        return self._valid


class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):

    def __init__(self, optimizers, model_components=None, batch_chunk_size=None):
        """
        Simple trainer based on a training_model, that evaluates the loss on batch data

        :param optimizers: dict or list with optimizer(s), or a single optimizer instance
        :param model_components: dict or list with model components(s), or a single model instance
        :param batch_chunk_size: optional batch chunk size (int)
                                 If given, batches are processed in chunks of size `batch_chunk_size` samples to
                                 calculate the gradients. The last chunk can be smaller than `batch_chunk_size` if
                                 there is not an exact multiple that is equal to the `batch_data` size

                                 Note 1.
                                 Chunked processing of a batch only works when the `batch_data` object, received
                                 by the `train_on` method, implements the `__len__` and `__getitem__` methods.
                                 Here the `__getitem__` method must be able to deal with slices.

                                 Note 2.
                                 When using chunked batch processing, the default implementation assumes that the
                                 loss, calculated over a chunk, is the average of the sample losses
        """

        model_components = convert_to_dict("model", model_components)
        optimizers = convert_to_dict("optimizer", optimizers)

        super().__init__(model_components, optimizers)

        self.batch_chunk_size = batch_chunk_size
        if self.batch_chunk_size is not None:
            self._log.info(f"Will train on batches by slicing the batch is chunks of {batch_chunk_size} samples.")

        self.training_model = None

    def set_training_model(self, model):
        """
        :param model: nn.Module that returns the loss based on the given batch
                      The forward method of the training model must be callable and the following signature:
                      model(self, batch_data, training_settings) -> (loss, auxiliary_results)

        :param model:
        :return:
        """

        self.training_model = model

        instance_valid = self.instance_valid()
        self._valid = self._validate_model() & instance_valid

    def evaluate_loss(self, batch_data, inference_mode, evaluate_settings=None):
        """

        :param batch_data: batch_data object to evaluate loss on (e.g. dict, list, tuple)
        :param inference_mode: If True the loss will be evaluated in inference mode (e.g. no Dropout).
                               If False the loss will be evaluated in training mode
        :param evaluate_settings: optional evaluate_settings object (usually dict)

        :return: loss, auxiliary_results

        loss : Tensor
        auxiliary_results : can be anything, e.g dict or list with values or data items
        """

        self._activate_inference_mode(inference_mode)

        return self._evaluate_loss(batch_data, evaluate_settings)

    def _validate_model(self):
        # TODO : this is framework dependent
        model_valid = callable(self.training_model)
        if not model_valid:
            self._log.error(f"Given training model {self.training_model} is invalid, trainer will not work")

        return model_valid
